structure DataTypes =
struct
datatype
         id = id of string
and      
         csx_type = Int | Bool | Char | Void | StringLit of int | UnknownConst 
                  | NoneType 
                  | Error (* used in the typechecker. *)
                  | Const of csx_type
                  | Param of csx_type
                  | Array of (csx_type * Int32.int)
                  | ArrayParam of csx_type
and
         arithmetic_operator = PLUS | MINUS | TIMES | DIV 
and      
         logical_operator = AND | OR 
and
         relational_operator = LT | GT | LEQ | GEQ 
and    
         equality_operator = EQ | NOTEQ
and
         unary_op = NOT
                  | Cast of csx_type
and 
         label = Label of id
               | LabelBeginEnd of (id * string * string)
               | NoneLabel
and
         expr = NoneExpr
              | ArithmeticBinaryOperation of (expr * arithmetic_operator * expr * int * int)
              | LogicalBinaryOperation of (expr * logical_operator * expr * int * int)
              | RelationalBinaryOperation of (expr * relational_operator * expr * int * int)
              | EqualityBinaryOperation of (expr * equality_operator * expr * int * int)
              | UnaryOperation of  (unary_op * expr * int * int)
              | CharLit of (string * int * int)
              | IntLit of (Int32.int * int * int)
              | StrLit of (string * int * int)
              | True of (int * int)
              | False of (int * int)
              | ArglessFnCall of (id * int * int)
              | FnCall of (id * expr list * int * int)
              | Name of (id * int * int)
              | ArrayDeref of (id * expr * int * int)
              | ActualArgument of (expr * int * int)
              | Checked of (csx_type * expr)
and
         csx_decl = Field of (csx_type * id * int * int)
                  | InitializedField of (csx_type * id * expr * int * int)
                  | Method of (csx_type * id * csx_decl list * csx_decl list * stmt list * int * int)
                  | MainMethod of (csx_decl list * stmt list * int * int)
                  | FormalArg of (csx_type * id * int * int)
                  | NoneDecl
and 
         csx_member_decls = MemberDecls of (csx_decl list * csx_decl list)
and      
         stmt = NoneStmt
              | Block of (csx_decl list * stmt list * int * int)
              | IfThen of (expr * stmt * int * int)
              | IfThenElse of (expr * stmt * stmt * int * int)
              | While of (expr * stmt * int * int)
              | LabeledWhile of (label * expr * stmt * int * int)
              | Asg of (expr * expr * int * int)
              | Read of (expr * int * int)
              | Reads of (stmt list * int * int)
              | Write of (expr * int * int)
              | Writes of (stmt list * int * int)
              | ArglessProcCall of (id * int * int)
              | ProcCall of (id * expr list * int * int)
              | Return of (expr * int * int)
              | Break of (label * int * int)
              | Continue of (label * int * int)
              | Exit of (Int32.int * int * int)
and 
         csx_program = Program of (id * csx_member_decls)

fun typeToString(Int) = "Int"
  | typeToString(Bool) = "Bool"
  | typeToString(Char) = "Char"
  | typeToString(Void) = "Void"
  | typeToString(UnknownConst) = "UnknownConst"
  | typeToString(Const(t)) = "Const("^typeToString(t)^")"
  | typeToString(Error) = "Error"
  | typeToString(Param(t)) = typeToString(t)^"Param"
  | typeToString(Array(t, n)) = typeToString(t)^"["^Int32.toString(n)^"]"
  | typeToString(ArrayParam(t)) = typeToString(t)^"[]Param"
  | typeToString(StringLit(n)) = "StringLit"
  | typeToString(NoneType) = "NoneType" 


exception Bug;
fun jvm_type(Int) = "I"
  | jvm_type(Bool) = "Z"
  | jvm_type(Char) = "C"
  | jvm_type(Void) = "V"
  | jvm_type(Const(t)) = jvm_type(t)
  | jvm_type(Param(t)) = jvm_type(t)
  | jvm_type(Array(t, n)) = "["^jvm_type(t)
  | jvm_type(ArrayParam(t)) = "["^jvm_type(t)
  | jvm_type(UnknownConst) = raise Bug 
  | jvm_type(Error) = raise Bug 
  | jvm_type(StringLit(n)) = raise Bug 
  | jvm_type(NoneType) = raise Bug


fun jasmin_type(Bool) = "boolean"
  | jasmin_type(Char) = "char"
  | jasmin_type(Int) = "int"
  | jasmin_type(_) = raise Bug 

fun method_args(Method(tipe, id(s), args, fields, body, _, _)) = args
  | method_args(_) = []

fun method_return_type(Method(tipe, id(s), args, fields, body, _, _)) = tipe
  | method_return_type(MainMethod(_, _, _, _)) = Void 
  | method_return_type(_) = NoneType 

  
end;
