structure PushCount =
struct 
open DataTypes;

  fun reduce(f, init, coll) = List.foldr (f) init coll;
  fun sum(ints) = reduce((op +), 0, ints);

fun push_count_stmts(stmts) = 
        sum(List.map (fn s => stmt_push_count(s)) stmts)
and 
       stmt_push_count(NoneStmt) = 0
     | stmt_push_count(Block(decls, stmts, _, _)) = 
            push_count_stmts(stmts)
     | stmt_push_count(IfThen(cond, then_stmt, _, _)) = 
            stmt_push_count(then_stmt)
     | stmt_push_count(IfThenElse(cond, then_stmt, else_stmt, _, _)) = 
            stmt_push_count(then_stmt) + stmt_push_count(else_stmt)
     | stmt_push_count(While(cond, stmt, _, _)) = 
            stmt_push_count(stmt)
     | stmt_push_count(LabeledWhile(_, cond, stmt, _, _)) = 
            stmt_push_count(stmt)
     | stmt_push_count(Asg(lval, rval, _, _)) =
            expr_push_count(lval) + expr_push_count(rval)
     | stmt_push_count(Read(expr, _, _)) = 
            expr_push_count(expr) 
     | stmt_push_count(Reads(reads, _, _)) = 
            push_count_stmts(reads)
     | stmt_push_count(Writes(writes, _, _)) = 
            push_count_stmts(writes)
     | stmt_push_count(Write(expr, _, _)) = expr_push_count(expr)
     | stmt_push_count(ArglessProcCall(id(name), _, _)) = 1
     | stmt_push_count(ProcCall(id(name), args, _, _)) = 
            1 + List.length(args)
     | stmt_push_count(Return(NoneExpr, _, _)) = 0
     | stmt_push_count(Return(expr, _, _)) = 
            expr_push_count(expr) 
     | stmt_push_count(Break(_, _, _)) = 0
     | stmt_push_count(Continue(_, _, _)) = 0
     | stmt_push_count(Exit(code, _, _)) = 1

and
       exprs_push_count(exprs) = 
        List.foldr (op +) 0 (List.map (fn e => expr_push_count(e)) exprs)
and
       expr_push_count(NoneExpr) = 0
     | expr_push_count(ArithmeticBinaryOperation(lhs, arit_binary_operator, rhs, _, _)) = 
            expr_push_count(lhs) + expr_push_count(rhs)
     | expr_push_count(LogicalBinaryOperation(lhs, logi_binary_operator, rhs, _, _)) =
            expr_push_count(lhs) + expr_push_count(rhs)
     | expr_push_count(RelationalBinaryOperation(lhs, rel_binary_operator, rhs, _, _)) =
            expr_push_count(lhs) + expr_push_count(rhs)
     | expr_push_count(EqualityBinaryOperation(lhs, equal_binary_operator, rhs, _, _)) =
            expr_push_count(lhs) + expr_push_count(rhs)
     | expr_push_count(UnaryOperation (unary_op, expr, _, _)) =
            expr_push_count(expr)
     | expr_push_count(CharLit(the_char, _, _)) = 1
     | expr_push_count(IntLit(the_num, _, _)) = 1
     | expr_push_count(StrLit(the_str, _, _)) = 1
     | expr_push_count(True(_)) = 1
     | expr_push_count(False(_)) = 1
     | expr_push_count(ArglessFnCall(id(name), _, _)) = 1
     | expr_push_count(FnCall(id(name), actual_args, _, _)) = 
            1 + exprs_push_count(actual_args)
     | expr_push_count(Name(id(name), _, _)) = 1
     | expr_push_count(ArrayDeref(id(name), expr, _, _)) = 1 + expr_push_count(expr)
     | expr_push_count(ActualArgument(expr, _, _)) = expr_push_count(expr)
     | expr_push_count(Checked(tipe, expr)) = expr_push_count(expr)

fun say(Program(id(name), members)) = ()
and 
    push_count_members(MemberDecls(fields, methods)) = push_count_decls(fields) + push_count_decls(methods)
and 
    push_count_actual_args(args) = ()
and 
    push_count_decls(decls) = sum (List.map (fn d => push_count_decl(d)) decls)
and
      push_count_decl(Field(Array(tipe, size), id(name), _, _)) = 2
    | push_count_decl(Field(tipe, id(s), _, _)) = 1
    | push_count_decl(InitializedField(tipe, id(s), value, _, _)) = 1 + expr_push_count(value)
    | push_count_decl(FormalArg(ArrayParam(tipe), id(name), _, _)) = 2
    | push_count_decl(FormalArg(tipe, id(name), _, _)) = 1
    | push_count_decl(Method(tipe, id(name), arg_decls, field_decls, stmts, _, _)) = 
       push_count_decls(arg_decls) + push_count_decls(field_decls) + push_count_stmts(stmts)
    | push_count_decl(MainMethod(field_decls, stmts, _, _)) = 
       push_count_decls(field_decls) + push_count_stmts(stmts)
    | push_count_decl(NoneDecl) = 0
end;
